import joblib
import numpy
import os
import tenseal


class CKKSTensor:
    context = tenseal.context_from(
        joblib.load(os.getenv('HEFLOW_CKKS', 'id_ckks')))

    def __getstate__(self):
        return {'data': self.backend.serialize()}

    def __init__(self, data):
        self.backend = tenseal.ckks_tensor(self.context, data)

    def __setstate__(self, state):
        self.backend = tenseal.ckks_tensor_from(self.context, state['data'])

    def add(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        return CKKSTensor(self.backend.add(other))

    def add_(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        self.backend.add_(other)
        return self

    def broadcast_to(self, shape):
        return CKKSTensor(self.backend.broadcast(shape))

    def broadcast_to_(self, shape):
        self.backend.broadcast_(shape)
        return self

    def clone(self):
        return CKKSTensor(self.backend.copy())

    def dot(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        return CKKSTensor(self.backend.dot(other))

    def dot_(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        self.backend.dot_(other)
        return self

    @property
    def dtype(self):
        return 'float64'

    def mm(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        return CKKSTensor(self.backend.mm(other))

    def mm_(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        self.backend.mm_(other)
        return self

    def mul(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        return CKKSTensor(self.backend.mul(other))

    def mul_(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        self.backend.mul_(other)
        return self

    def neg(self):
        return CKKSTensor(self.backend.neg())

    def neg_(self):
        self.backend.neg_()
        return self

    def numpy(self):
        backend = self.backend.decrypt()
        return numpy.asarray(backend.raw, backend.dtype).reshape(backend.shape)

    def polyval(self, coeffs):
        return CKKSTensor(self.backend.polyval(coeffs))

    def polyval_(self, coeffs):
        self.backend.polyval_(coeffs)
        return self

    def pow(self, exponent):
        return CKKSTensor(self.backend.pow(exponent))

    def pow_(self, exponent):
        self.backend.pow_(exponent)
        return self

    def reshape(self, shape):
        return CKKSTensor(self.backend.reshape(shape))

    def reshape_(self, shape):
        self.backend.reshape_(shape)
        return self

    @property
    def shape(self):
        return self.backend.shape

    def square(self):
        return CKKSTensor(self.backend.square())

    def square_(self):
        self.backend.square_()
        return self

    def sub(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        return CKKSTensor(self.backend.sub(other))

    def sub_(self, other):
        if isinstance(other, CKKSTensor):
            other = other.backend

        self.backend.sub_(other)
        return self

    def sum(self):
        return CKKSTensor(self.backend.sum())

    def sum_(self):
        self.backend.sum_()
        return self

    def transpose(self):
        return CKKSTensor(self.backend.transpose())

    def transpose_(self):
        self.backend.transpose_()
        return self


def ckks_tensor(data):
    return CKKSTensor(data)



import heflow.sklearn
import mlflow.sklearn
import sklearn.datasets
import sklearn.linear_model
import sklearn.metrics
import sklearn.model_selection

X, y = sklearn.datasets.fetch_openml('iris', return_X_y=True)

X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
    X, y, test_size=3, random_state=42)

mlflow.sklearn.autolog(log_models=False)

with mlflow.start_run():
    lr = sklearn.linear_model.LogisticRegression()

    model = lr.fit(X_train, y_train)

    predictions = model.predict(X_test)

    print('predictions=%s' % predictions)

    print('accuracy=%.3f' %
          sklearn.metrics.accuracy_score(y_test, predictions))

    heflow.sklearn.log_model(model)



import heflow.models
import heflow.tensors
import numpy


@heflow.models.ckks_model('decision_function')
class LogisticRegression:

    def __init__(self, model):
        self.classes_ = model.classes_
        self.he_coef_ = heflow.tensors.ckks_tensor(model.coef_).transpose_()
        self.he_intercept_ = heflow.tensors.ckks_tensor(model.intercept_)

    def decision_function(self, he_X):
        he_scores = he_X.dot_(self.he_coef_).add_(self.he_intercept_)
        return he_scores.reshape_(
            (he_scores.shape[0], )) if he_scores.shape[1] == 1 else he_scores

    def he_predict(self, X):
        he_X = heflow.tensors.ckks_tensor(X)
        he_scores = self.decision_function(he_X)
        scores = he_scores.numpy()
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = numpy.argmax(scores, axis=1)
        return numpy.take(self.classes_, indices, axis=0)



import click
import joblib
import json
import pathlib
import tenseal


def jsondict(ctx, param, value):
    if isinstance(value, tuple):
        options = value
    else:
        options = (value, )
    d = {}
    for option in options:
        key, value = option.split('=', 1)
        d[key] = json.loads(value)
    return d


@click.command(options_metavar='[options] — HEflow authentication key utility')
@click.option('-f',
              help='Specifies the filename of the key file.',
              type=pathlib.Path,
              metavar='filename')
@click.option(
    '-O',
    multiple=True,
    help=
    'Specify a key/value option.  These are specific to the operation that heflow-keygen has been requested to perform.  The -O option may be specified multiple times.',
    callback=jsondict,
    metavar='option')
@click.option(
    '-t',
    help=
    'Specifies the type of key to create.  The possible values are "ckks".',
    default='ckks',
    metavar='ckks')
@click.option(
    '-y',
    is_flag=True,
    help=
    'This option will read a private HEflow format file and print an HEflow public key to stdout.',
    default=False,
    metavar='')
@click.option('-q', help='Silence heflow-keygen.', default=False, metavar='')
@click.pass_context
def command(ctx, f, o, t, y, q):
    """
    heflow-keygen [-b bits] [-f output_keyfile] [-O option] [-t ckks]

    heflow-keygen -y [-f input_keyfile]
    """
    if y:
        if not f:
            f = click.prompt('Enter file in which the key is',
                             'id_ckks',
                             type=click.Path(path_type=pathlib.Path))
        try:
            context = tenseal.context_from(
                joblib.load(click.open_file(f, 'rb')))
        except OSError as ose:
            click.echo(f'{f}: {ose.strerror}', err=True)
            ctx.exit(255)
        joblib.dump(context.serialize(), click.get_binary_stream('stdout'))
    elif t == 'ckks':
        if not q:
            click.echo(f'Generating public/private {t} key pair.')
        context = tenseal.context(tenseal.SCHEME_TYPE.CKKS,
                                  o.get('poly_modulus_degree', 16384),
                                  coeff_mod_bit_sizes=o.get(
                                      'coeff_modulus_bit_sizes',
                                      [52, 52, 52, 52, 52, 52, 52]))
        context.global_scale = 2**o.get('scale_bit_size', 52)
        context.generate_galois_keys()
        if not f:
            f = click.prompt('Enter file in which to save the key',
                             'id_ckks',
                             type=click.Path(path_type=pathlib.Path))
        if f.exists():
            click.echo(f'{f} already exists.')
            if click.prompt(
                    'Overwrite (y/n)', '', prompt_suffix='? ',
                    show_default=False) != 'y':
                ctx.exit(1)
        try:
            joblib.dump(context.serialize(save_secret_key=True), f)
        except OSError as ose:
            click.echo(f'Saving key "{f}" failed: {ose.strerror}', err=True)
            ctx.exit(255)
        if not q:
            click.echo(f'Your identification has been saved in {f}')
        try:
            joblib.dump(context.serialize(), f'{f}.pub')
        except OSError as ose:
            click.echo(f'Unable to save public key to {f}.pub: {ose.strerror}',
                       err=True)
            ctx.exit(255)
        if not q:
            click.echo(f'Your public key has been saved in {f}.pub')
    else:
        click.echo(f'unknown key type {t}', err=True)


if __name__ == '__main__':
    command()



import heflow.tensors
import mlserver.codecs
import mlserver.codecs.utils
import mlserver.types
import pickle
import typing


@mlserver.codecs.register_input_codec
class CKKSTensorCodec(mlserver.codecs.InputCodec):
    ContentType: typing.ClassVar[str] = 'heflow/ckks_tensor'
    TypeHint: typing.ClassVar[typing.Type] = heflow.tensors.CKKSTensor

    @classmethod
    def can_encode(cls, payload: typing.Any) -> bool:
        return isinstance(payload, heflow.tensors.CKKSTensor)

    @classmethod
    def encode_output(cls, name: str, payload: heflow.tensors.CKKSTensor,
                      **kwargs) -> mlserver.types.ResponseOutput:
        return mlserver.types.ResponseOutput(
            name=name,
            shape=payload.shape,
            datatype='BYTES',
            parameters=mlserver.types.Parameters(content_type=cls.ContentType),
            data=[pickle.dumps(payload)],
        )

    @classmethod
    def decode_output(
        cls, response_output: mlserver.types.ResponseOutput
    ) -> heflow.tensors.CKKSTensor:
        return pickle.loads(response_output.data[0])

    @classmethod
    def encode_input(cls, name: str, payload: heflow.tensors.CKKSTensor,
                     **kwargs) -> mlserver.types.RequestInput:
        return mlserver.types.RequestInput(
            name=name,
            shape=payload.shape,
            datatype='BYTES',
            parameters=mlserver.types.Parameters(content_type=cls.ContentType),
            data=[pickle.dumps(payload)],
        )

    @classmethod
    def decode_input(
        cls, request_input: mlserver.types.RequestInput
    ) -> heflow.tensors.CKKSTensor:
        return pickle.loads(request_input.data[0])


class InputRequestCodec(mlserver.codecs.RequestCodec):
    InputCodec: typing.Optional[mlserver.codecs.InputCodecLike] = None

    @classmethod
    def can_encode(cls, payload: typing.Any) -> bool:
        if cls.InputCodec is None:
            return False

        if isinstance(payload, tuple):
            payloads = payload
        else:
            payloads = (payload, )
        for payload in payloads:
            if not cls.InputCodec.can_encode(payload):
                return False
        return True

    @classmethod
    def encode_response(
        cls,
        model_name: str,
        payload: typing.Any,
        model_version: typing.Optional[str] = None,
        **kwargs,
    ) -> mlserver.types.InferenceResponse:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f'No input codec found for {type(cls)} request codec')

        outputs = []
        if isinstance(payload, tuple):
            payloads = payload
        else:
            payloads = (payload, )
        for index, payload in enumerate(payloads, 1):
            outputs.append(
                cls.InputCodec.encode_output(
                    f'{mlserver.codecs.utils.DefaultOutputPrefix}{index}',
                    payload, **kwargs))
        return mlserver.types.InferenceResponse(
            model_name=model_name,
            model_version=model_version,
            parameters=mlserver.types.Parameters(content_type=cls.ContentType),
            outputs=outputs)

    @classmethod
    def decode_response(
            cls, response: mlserver.types.InferenceResponse) -> typing.Any:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f'No input codec found for {type(cls)} request codec')

        payloads = ()
        for output in response.outputs:
            payloads += (cls.InputCodec.decode_output(output), )
        if len(payloads) == 1:
            return payloads[0]
        return payloads

    @classmethod
    def encode_request(cls, payload: typing.Any,
                       **kwargs) -> mlserver.types.InferenceRequest:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f'No input codec found for {type(cls)} request codec')

        if isinstance(payload, tuple):
            payloads = payload
        else:
            payloads = (payload, )
        inputs = []
        for index, payload in enumerate(payloads, 1):
            inputs.append(
                cls.InputCodec.encode_input(
                    f'{mlserver.codecs.utils.DefaultInputPrefix}{index}',
                    payload, **kwargs))
        return mlserver.types.InferenceRequest(
            parameters=mlserver.types.Parameters(content_type=cls.ContentType),
            inputs=inputs)

    @classmethod
    def decode_request(cls,
                       request: mlserver.types.InferenceRequest) -> typing.Any:
        if cls.InputCodec is None:
            raise NotImplementedError(
                f'No input codec found for {type(cls)} request codec')

        payloads = ()
        for input in request.inputs:
            payloads += (cls.InputCodec.decode_input(input), )
        if len(payloads) == 1:
            return payloads[0]
        return payloads


@mlserver.codecs.register_request_codec
class CKKSTensorRequestCodec(InputRequestCodec):
    InputCodec: typing.Optional[
        mlserver.codecs.InputCodecLike] = CKKSTensorCodec
    ContentType: typing.ClassVar[str] = InputCodec.ContentType
    TypeHint: typing.ClassVar[typing.Type] = InputCodec.TypeHint



import heflow
import heflow.sklearn.linear_model
import sklearn.linear_model


def log_model(model):
    if isinstance(model, sklearn.linear_model.LogisticRegression):
        return heflow.log_model(
            heflow.sklearn.linear_model.LogisticRegression(model))

    raise TypeError(f'invalid model instance: {type(model)}')


def save_model(path, model):
    if isinstance(model, sklearn.linear_model.LogisticRegression):
        return heflow.save_model(
            path, heflow.sklearn.linear_model.LogisticRegression(model))

    raise TypeError(f'invalid model instance: {type(model)}')



from setuptools import find_packages, setup

setup(
    name='heflow',
    use_scm_version=True,
    description='HEflow: A platform for the homomorphic encryption lifecycle',
    long_description_content_type='text/markdown',
    author='InAccel',
    author_email='info@inaccel.com',
    url='https://github.com/inaccel/heflow',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    license='Apache-2.0',
    platforms=[
        'Linux',
    ],
    install_requires=[
        'mlflow[extras]',
        'tenseal',
    ],
    entry_points={
        'console_scripts': [
            'heflow-keygen = heflow.cli.keygen:command',
        ],
    },
    python_requires='>=3.8',
)



import heflow.codecs
import mlflow.pyfunc
import mlflow.pyfunc.mlserver
import mlserver.grpc.converters
import mlserver.grpc.dataplane_pb2_grpc


def ckks_model(infer: str):

    def decorate(cls):
        setattr(cls, '__call__', getattr(cls, infer))

        def connect(self, channel):
            setattr(self, 'channel', channel)
            return self

        setattr(cls, 'connect', connect)

        def disconnect(self):
            if hasattr(self, 'channel'):
                delattr(self, 'channel')
            return self

        setattr(cls, 'disconnect', disconnect)

        def stub(self, *args):
            if not hasattr(self, 'channel'):
                return self(*args)

            return heflow.codecs.CKKSTensorRequestCodec.decode_response(
                mlserver.grpc.converters.ModelInferResponseConverter.to_types(
                    mlserver.grpc.dataplane_pb2_grpc.GRPCInferenceServiceStub(
                        self.channel).ModelInfer(
                            mlserver.grpc.converters.
                            ModelInferRequestConverter.from_types(
                                heflow.codecs.CKKSTensorRequestCodec.
                                encode_request(args),
                                model_name=mlflow.pyfunc.mlserver.
                                MLServerDefaultModelName))))

        setattr(cls, infer, stub)

        def predict(self, context, model_input):
            if isinstance(model_input, tuple):
                return self(*model_input)
            else:
                return self(model_input)

        return type('CKKSModel', (cls, mlflow.pyfunc.PythonModel),
                    {'predict': predict})

    return decorate



import mlflow.pyfunc


def load_model(model_uri: str):
    return mlflow.pyfunc.load_model(model_uri).unwrap_python_model()


def log_model(model):
    return mlflow.pyfunc.log_model('model', python_model=model)


def save_model(path, model):
    return mlflow.pyfunc.save_model(path, python_model=model)
